import pytest
import requests

from stelar.client import *
from stelar.client.tool import ToolProfile


def create_test_tool(testcli: Client) -> Tool:
    tool = testcli.tools.create(
        name="test-tool",
        description="A tool for testing purposes",
        parameters={
            "x": "The x coordinate",
            "y": "The y coordinate",
            "z": "The z coordinate",
        },
        inputs={
            "map_file": "The file containing the map data",
            "vec_file": "The file containing the vector data",
        },
        outputs={
            "result": "The result of the tool execution",
            "log": "The log file of the tool execution",
        },
        profiles={
            "default": {
                "image": "v0.8",
                "memory_limit": "2G",
            },
            "large": {
                "image": "v0.8",
                "memory_limit": "4G",
            },
        },
    )
    return tool


@pytest.fixture
def test_tool(testcli):
    tool = create_test_tool(testcli)

    yield tool

    tool.delete()


def test_tool_creation(testcli, test_tool: Tool):
    tool = test_tool
    assert isinstance(tool, Tool), "The created object is not a Tool instance"

    try:
        assert tool.name == "test-tool"
        assert tool.description == "A tool for testing purposes"
        assert tool.parameters == {
            "x": "The x coordinate",
            "y": "The y coordinate",
            "z": "The z coordinate",
        }
        assert tool.inputs == {
            "map_file": "The file containing the map data",
            "vec_file": "The file containing the vector data",
        }
        assert tool.outputs == {
            "result": "The result of the tool execution",
            "log": "The log file of the tool execution",
        }
        assert tool.profiles == {
            "default": {
                "image": "v0.8",
                "memory_limit": "2G",
            },
            "large": {
                "image": "v0.8",
                "memory_limit": "4G",
            },
        }
    finally:
        tool.delete()


def test_tool_in_list(testcli, test_tool: Tool):
    assert "test-tool" in testcli.tools
    assert test_tool is testcli.tools["test-tool"]


def test_tool_category(test_tool):
    assert test_tool.category is None

    test_tool.category = "other"

    assert test_tool.category == "other"

    with pytest.raises(ValueError):
        test_tool.category = "invalid category"


def test_tool_repository(test_tool):
    assert test_tool.repository == test_tool.name


def test_tool_registry_access(testcli, test_tool: Tool):
    url = testcli.klms_info.klms_root_url.replace("klms.", "registry.")
    headers = {"Authorization": f"Bearer {testcli._token}"}
    response = requests.get(
        f"{url}/api/v1/repository/stelar/{test_tool.name}", headers=headers
    )
    assert (
        response.status_code == 200
    ), f"Failed to access the tool registry: {response.text}"
    repo = response.json()

    assert (
        repo["name"] == test_tool.name
    ), "The repository name does not match the tool name"


def test_tool_no_profiles(testcli):
    tool = testcli.tools.create(
        name="test-tool-no-profiles",
        description="A tool without profiles",
        parameters={},
        inputs={},
        outputs={},
    )

    try:
        assert tool.profiles == {}

        assert tool.get_profile() is None
        assert tool.get_profile("default") is None

        with pytest.raises(ValueError):
            tool.profiles = None

        tool.set_profile("default")
        assert tool.profiles["default"] == {}

        tool.set_profile("default", image="v0.8", memory_limit="2G")
        assert tool.profiles["default"] == {
            "image": "v0.8",
            "memory_limit": "2G",
        }

        tool.update_profile("default", image="v0.9")
        assert tool.profiles["default"] == {
            "image": "v0.9",
            "memory_limit": "2G",
        }

        tool.set_profile("default", memory_limit="3G")
        assert tool.profiles["default"] == {
            "memory_limit": "3G",
        }

        tool.delete_profile("default")
        assert "default" not in tool.profiles

    finally:
        tool.delete()


def test_tool_with_profiles(testcli, test_tool):
    tool = test_tool

    assert tool.profiles
    assert tool.get_profile("default") == ToolProfile(image="v0.8", memory_limit="2G")
    assert tool.get_profile("large") == ToolProfile(image="v0.8", memory_limit="4G")

    assert len(tool.profiles) == 2
    tool.delete_profile("default")


def test_tool_get_image(testcli):
    t = testcli.tools["simple_tool"]

    assert t.get_image().name == "0.2.0"
    assert t.get_image("0.2.0").name == "0.2.0"

    assert t.get_image("1.2.0") is None
    assert t.get_image("1.2.0", "v0.8") == "v0.8"



from __future__ import annotations

import re
from dataclasses import dataclass, fields
from datetime import datetime
from typing import Any

from frozendict import deepfreeze, frozendict

from .api_call import api_call
from .generic import GenericCursor, GenericProxy
from .license import LicensedProxy
from .package import PackageCursor, PackageProxy
from .proxy import (
    AnyField,
    DateFmtField,
    EnumeratedField,
    Id,
    Property,
    StrField,
    derived_property,
)
from .proxy.property import DictProperty
from .task_spec import TaskSpec


class ToolCategoryField(EnumeratedField):
    VALUES = [
        "discovery",
        "interlinking",
        "annotation",
        "other",
    ]


#    {'is_manifest_list': False,                                            │(stelar) vsam@metallurgix:~/git/stelar/stelar_client$ git commit -a
#    'last_modified': 'Sun, 01 Jun 2025 19:45:45 -0000',                             │[main e7923b9] Fixed the proxy resync code related to task creation and termination.
#    'manifest_digest': 'sha256:6e4a55d6531c1aa8844164b552ba7b5f5ef9b30c5d9d94a145777│ 4 files changed, 49 insertions(+), 4 deletions(-)
# c7d26b7c572',                                                                       │(stelar) vsam@metallurgix:~/git/stelar/stelar_client$
#    'name': '0.2.0',                                                                │(stelar) vsam@metallurgix:~/git/stelar/stelar_client$
#    'reversion': False,                                                             │(stelar) vsam@metallurgix:~/git/stelar/stelar_client$ git pull
#    'size': 56837148,                                                               │Already up to date.
#    'start_ts': 1748807145
# }


@dataclass
class Image:
    """A container image used by a tool.

    Images are just information objects, they are not proxies.
    """

    is_manifest_list: bool
    last_modified: datetime
    manifest_digest: str
    tool: Tool  # The tool this image belongs to
    name: str
    reversion: bool
    size: int
    start_ts: datetime

    def __post_init__(self):
        if isinstance(self.last_modified, str):
            self.last_modified = datetime.strptime(
                self.last_modified, "%a, %d %b %Y %H:%M:%S -0000"
            )
        if isinstance(self.start_ts, int):
            self.start_ts = datetime.fromtimestamp(self.start_ts)

    @property
    def id(self):
        """Return a unique identifier for this image."""
        return f"stelar/{self.tool.name}:{self.name}"

    def __repr__(self):
        return f"Image({self.tool.name}/{self.name})"


IMAGE_TAG_PATTERN = re.compile(r"[a-zA-Z0-9_][a-zA-Z0-9_.-]{0,127}$")
KUBERNETES_NAME_PATTERN = re.compile(r"[a-z0-9]([-a-z0-9]*[a-z0-9])?")
CPU_QUANTITY_PATTERN = re.compile(r"^\d+(\.\d+)?[m]?$")
MEMORY_QUANTITY_PATTERN = re.compile(r"^\d+(\.\d+)?(Ki|Mi|Gi|Ti|Pi|Ei|k|M|G|T|P|E)?$")
RESTART_POLICY_VALUES = ["Always", "OnFailure", "Never"]
IMAGE_PULL_POLICY_VALUES = ["Always", "IfNotPresent", "Never"]


@dataclass(frozen=True)
class ToolProfile:
    """An object specifying options for executing images of a tool."""

    image: str | None = None
    """The image to use for this tool, if specified."""
    description: str | None = None
    image_pull_policy: str | None = None
    image_pull_secrets: list[str] | None = None
    cpu_request: str | None = None
    cpu_limit: str | None = None
    memory_request: str | None = None
    memory_limit: str | None = None
    backoff_limit: int | None = None
    restart_policy: str | None = None
    ttl_seconds_after_finished: int | None = None

    def __post_init__(self):
        self.validate_image(self.image)
        self.validate_image_pull_policy(self.image_pull_policy)
        self.validate_image_pull_secrets(self.image_pull_secrets)
        self.validate_cpu_request(self.cpu_request)
        self.validate_cpu_limit(self.cpu_limit)
        self.validate_memory_request(self.memory_request)
        self.validate_memory_limit(self.memory_limit)
        self.validate_backoff_limit(self.backoff_limit)
        self.validate_restart_policy(self.restart_policy)
        self.validate_ttl_seconds_after_finished(self.ttl_seconds_after_finished)

    @staticmethod
    def validate(attr_name: str, value: Any):
        """Validate a single attribute of the tool profile."""
        validator = getattr(ToolProfile, f"validate_{attr_name}", None)
        if validator is None:
            raise ValueError(f"Invalid attribute: {attr_name}")
        return validator(value)

    @staticmethod
    def validate_image(image) -> None:
        if image is not None and not IMAGE_TAG_PATTERN.fullmatch(image):
            raise ValueError(
                f"Invalid image name: {image}. Must match {IMAGE_TAG_PATTERN.pattern}"
            )

    @staticmethod
    def validate_image_pull_policy(policy: str | None) -> None:
        if policy is not None and policy not in IMAGE_PULL_POLICY_VALUES:
            raise ValueError(
                f"Invalid image pull policy: {policy}. Must be one of {IMAGE_PULL_POLICY_VALUES}."
            )

    @staticmethod
    def validate_image_pull_secrets(secrets: list[str] | None) -> None:
        if secrets is not None:
            if not isinstance(secrets, list | tuple):
                raise TypeError(
                    f"Image pull secrets must be a list or tuple, got {type(secrets)}"
                )
            if not all(KUBERNETES_NAME_PATTERN.fullmatch(secret) for secret in secrets):
                raise ValueError(
                    f"Invalid image pull secret names: {secrets}. Must match {KUBERNETES_NAME_PATTERN.pattern}"
                )

    @staticmethod
    def validate_cpu_request(cpu_request: str | None) -> None:
        if cpu_request is not None and not CPU_QUANTITY_PATTERN.fullmatch(cpu_request):
            raise ValueError(
                f"Invalid CPU request: {cpu_request}. Must match {CPU_QUANTITY_PATTERN.pattern}"
            )

    @staticmethod
    def validate_cpu_limit(cpu_limit: str | None) -> None:
        if cpu_limit is not None and not CPU_QUANTITY_PATTERN.fullmatch(cpu_limit):
            raise ValueError(
                f"Invalid CPU limit: {cpu_limit}. Must match {CPU_QUANTITY_PATTERN.pattern}"
            )

    @staticmethod
    def validate_memory_request(memory_request: str | None) -> None:
        if memory_request is not None and not MEMORY_QUANTITY_PATTERN.fullmatch(
            memory_request
        ):
            raise ValueError(
                f"Invalid memory request: {memory_request}. Must match {MEMORY_QUANTITY_PATTERN.pattern}"
            )

    @staticmethod
    def validate_memory_limit(memory_limit: str | None) -> None:
        if memory_limit is not None and not MEMORY_QUANTITY_PATTERN.fullmatch(
            memory_limit
        ):
            raise ValueError(
                f"Invalid memory limit: {memory_limit}. Must match {MEMORY_QUANTITY_PATTERN.pattern}"
            )

    @staticmethod
    def validate_backoff_limit(backoff_limit: int | None) -> None:
        if backoff_limit is not None and (
            not isinstance(backoff_limit, int) or backoff_limit < 0
        ):
            raise ValueError(
                f"Invalid backoff limit: {backoff_limit}. Must be a non-negative integer."
            )

    @staticmethod
    def validate_restart_policy(restart_policy: str | None) -> None:
        if restart_policy is not None and restart_policy not in RESTART_POLICY_VALUES:
            raise ValueError(
                f"Invalid restart policy: {restart_policy}. Must be one of {RESTART_POLICY_VALUES}."
            )

    @staticmethod
    def validate_ttl_seconds_after_finished(ttl_seconds: int | None) -> None:
        if ttl_seconds is not None and (
            not isinstance(ttl_seconds, int) or ttl_seconds < 0
        ):
            raise ValueError(
                f"Invalid TTL seconds after finished: {ttl_seconds}. Must be a non-negative integer."
            )

    def to_dict(self) -> dict[str, Any]:
        """Convert the tool profile to a dictionary, omit None values."""
        return {
            field.name: getattr(self, field.name)
            for field in fields(self)
            if getattr(self, field.name) is not None
        }


class ProfilesField(AnyField):
    """A field for storing tool profiles as a dictionary."""

    def __init__(self, **kwargs):
        super().__init__(nullable=False, **kwargs)
        self.add_check(self.to_frozendict, 5)

    def to_frozendict(
        self, value: dict[str, ToolProfile | dict], **kwargs
    ) -> tuple[frozendict, bool]:
        """Convert a dictionary of ToolProfile objects to a frozendict."""
        if not isinstance(value, (dict, frozendict)):
            raise TypeError(f"Expected dict, got {type(value)}")

        fprof = {}
        for key, profile in value.items():
            if not isinstance(profile, (ToolProfile | dict | frozendict)):
                raise TypeError(
                    f"Expected profile dict or ToolProfile, got {type(profile)}"
                )
            try:
                ToolProfile.validate_image(key)
            except Exception as e:
                raise ValueError(f"Invalid profile name '{key}': {e}")
            if isinstance(profile, dict | frozendict):
                try:
                    ToolProfile(**profile)
                except Exception as e:
                    raise ValueError(f"Invalid profile data for '{key}': {e}")
                fprof[key] = profile
            else:
                fprof[key] = profile.to_dict()
        return deepfreeze(fprof), False

    def convert_to_proxy(self, value: Any, **kwargs) -> frozendict:
        return deepfreeze(value)


class Tool(PackageProxy, LicensedProxy):
    # weird ones
    # license_id = Property(validator=StrField(nullable=True), updatable=True)
    git_repository = Property(validator=StrField(nullable=True), updatable=True)
    programming_language = Property(validator=StrField(nullable=True), updatable=True)
    version = Property(
        validator=StrField(nullable=True, maximum_len=100), updatable=True
    )

    inputs = DictProperty(str, str, updatable=True)
    outputs = DictProperty(str, str, updatable=True)
    parameters = DictProperty(str, str, updatable=True)
    category = Property(validator=ToolCategoryField(nullable=True), updatable=True)
    repository = Property(validator=StrField(nullable=False), updatable=False)

    profiles = Property(validator=ProfilesField, updatable=True)

    def get_profile(self, name="default", default=None) -> ToolProfile | None:
        """Return the tool profile for the given name."""
        if name not in self.profiles:
            return default
        return ToolProfile(**self.profiles[name])

    def set_profile(
        self, name: str, profile: ToolProfile | dict | None = None, **pargs
    ) -> None:
        ps = self.profiles
        if profile is None:
            profile = ToolProfile(**pargs)
        elif isinstance(profile, dict):
            profile = ToolProfile(**(profile | pargs))

        self.profiles = ps.set(name, profile)

    def update_profile(self, name: str, **pargs) -> None:
        """Update the tool profile with the given name."""
        ps = self.profiles
        prof = ps.get(name, {})

        prof |= pargs
        return self.set_profile(name, prof)

    def delete_profile(self, name: str) -> None:
        """Delete the tool profile with the given name."""
        ps = dict(self.profiles)
        if name in ps:
            del ps[name]
            self.profiles = ps

    @derived_property
    def images(self, entity):
        """Return a list of images associated with this tool."""
        images = entity.get("images", [])
        return [Image(tool=self, **image) for image in images]

    def get_image(self, name: str | None = None, default=None) -> Image | None:
        """Return the image of the given name, or the default image

        for this tool, if any.
        """
        if name is not None:
            for image in self.images:
                if image.name == name:
                    return image
            return default

        # No name, return the default image if it exists
        dprof = self.get_profile()
        if dprof and dprof.image:
            # If a default profile is set, return the image from that profile
            return self.get_image(dprof.image, default)
        img = self.images
        if len(img) == 1:
            # If there's only one image, return it
            return img[0]
        return None

    def task_spec(self, profile_or_image=None):
        """Return a new TaskSpec initialized for this tool.

        Args
            profile_or_image: Either a profile name or an image name.
                If a profile name is given, the image from that profile
                will be used. If an image name is given, it will be used
                directly.
        """
        # if image not in self.profiles
        return TaskSpec(tool=self, image=profile_or_image)


class ToolCursor(PackageCursor[Tool]):
    def __init__(self, client):
        super().__init__(client, Tool)


class ImageRegistryToken(GenericProxy):
    """A token representing an image used by a tool."""

    DATEFORMAT = "%a, %d %b %Y %H:%M:%S %z"

    id = Id(entity_name="uuid")
    created = Property(validator=DateFmtField(DATEFORMAT), updatable=False)
    expiration = Property(
        validator=DateFmtField(DATEFORMAT, nullable=True), updatable=False
    )
    last_accessed = Property(
        validator=DateFmtField(DATEFORMAT, nullable=True), updatable=False
    )

    title = Property(validator=StrField(nullable=False), updatable=False)

    token = Property(
        validator=StrField(nullable=False), updatable=False, entity_name="token_code"
    )


class ImageRegistryTokenCursor(GenericCursor[ImageRegistryToken]):
    """A cursor for iterating over image registry tokens."""

    def __init__(self, client):
        super().__init__(client, ImageRegistryToken)

    def create(
        self, title: str, expiration: datetime | None = None
    ) -> ImageRegistryToken:
        """Create a new image registry token."""
        ac = api_call(self)
        result = ac.image_registry_token_create(title=title, expiration=expiration)
        return self.fetch_proxy_for_entity(result)



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Any
from uuid import UUID

from .api_call import api_call
from .generic import GenericCursor, GenericProxy, generic_proxy_sync
from .lineage import Lineage
from .mutils import is_s3url, s3spec_to_url
from .pdutils import read_dataframe
from .proxy import (
    DateField,
    ExtrasProperty,
    Id,
    IntField,
    Property,
    Proxy,
    ProxyVec,
    Reference,
    StrField,
)
from .proxy.refs import RefField
from .reprstyle import resource_to_html
from .utils import client_for

if TYPE_CHECKING:
    from .mutils import S3ObjSpec


class ExtrasResourceProperty(ExtrasProperty):
    """The handling of extras is different for Resource entities."""

    # The format is that extra attributes appear normally in
    # resource entities; there is no extras attribute.

    def convert_entity_to_proxy(self, proxy: Proxy, entity: Any, **kwargs):
        # to recognize extras fields, use the schema
        all_fields = proxy.proxy_schema.all_entity_fields
        proxy_extras = {p: v for p, v in entity.items() if p not in all_fields}
        proxy.proxy_attr[self.name] = proxy_extras

    def convert_proxy_to_entity(self, proxy: Proxy, entity: dict, **kwargs):
        proxy_extras = proxy.proxy_attr[self.name]
        entity |= proxy_extras

    def convert_to_create(
        self, proxy_type, create_props: dict, entity_props: dict, **kwargs
    ):
        all_fields = proxy_type.proxy_schema.all_fields

        # Collect all entries that do not appear in the schema
        entity_extras = {p: v for p, v in create_props.items() if p not in all_fields}
        entity_props |= entity_extras


class PackageField(RefField):
    """A field validator for package references."""

    @property
    def proxy_type(self):
        """Return the proxy type for the package reference."""
        return self.ref_property.proxy_types


class PackageRef(Reference):
    """A polymprphic reference to a package entity, used in Resource."""

    VALIDATOR_CLASS = PackageField

    # All entities with resources
    ENTITIES = ("Dataset", "Process", "Tool", "Workflow")

    def __init__(self):
        super().__init__(
            proxy_type="Dataset",
            nullable=False,
            trigger_sync=True,
            entity_name="package_id",
            updatable=False,
        )

    @cached_property
    def proxy_types(self):
        """Return the proxy types that this reference can point to."""
        from .proxy.schema import Schema

        return tuple(Schema.for_entity(t).cls for t in self.ENTITIES)

    def registry_for(self, obj) -> Any:
        """Return the registry for the package reference, given owner."""
        # Use the package_type attribute to determine the type of package.
        if (
            obj.proxy_attr
            and "_extras" in obj.proxy_attr
            and "package_type" in obj.proxy_attr["_extras"]
        ):
            package_type = obj.proxy_attr["_extras"].get("package_type", "dataset")
        else:
            package_type = "dataset"

        return obj.proxy_registry.catalog.registry_for(package_type.capitalize())


class Resource(GenericProxy):
    """
    A proxy for a STELAR resource with metadata and additional details.
    """

    id = Id()

    # dataset = Reference("Dataset", entity_name="package_id", trigger_sync=True)
    dataset = PackageRef()

    position = Property(validator=IntField)
    state = Property(validator=StrField)
    metadata_modified = Property(validator=DateField)

    url = Property(validator=StrField, updatable=True)
    format = Property(validator=StrField, updatable=True)
    description = Property(validator=StrField, updatable=True)
    hash = Property(validator=StrField, updatable=True)

    name = Property(validator=StrField, updatable=True)
    resource_type = Property(validator=StrField, updatable=True)
    mimetype = Property(validator=StrField, updatable=True)
    mimetype_inner = Property(validator=StrField, updatable=True)
    cache_url = Property(validator=StrField, updatable=True)
    size = Property(validator=IntField, updatable=True)
    created = Property(validator=DateField, updatable=True)
    last_modified = Property(validator=DateField, updatable=True)
    cache_last_updated = Property(validator=DateField, updatable=True)
    url_type = Property(validator=StrField, updatable=True)

    _extras = ExtrasResourceProperty()

    def __getattr__(self, attr):
        try:
            return self.proxy_schema.extras.get(self)[attr]
        except KeyError as e:
            raise AttributeError(attr) from e

    def __setattr__(self, attr, value):
        if attr.startswith("proxy_") or attr in self.proxy_schema.all_fields:
            return object.__setattr__(self, attr, value)

        prop = self.proxy_schema.extras

        # TODO: value validation: It is not clear what to do, presumably the correct
        # value would be transformable to json

        prop.touch(self)
        self.proxy_schema.extras.get(self)[attr] = value
        self.proxy_autocommit()

    def __delattr__(self, attr):
        if attr.startswith("proxy_") or attr in self.proxy_schema.all_fields:
            return object.__delattr__(self, attr)

        prop = self.proxy_schema.extras
        extras = prop.get(self)
        if attr in extras:
            prop.touch(self)
            del extras[attr]
            self.proxy_autocommit()
        else:
            raise AttributeError(attr)

    def proxy_sync(self, entity=None):
        return generic_proxy_sync(self, entity, update_method="update")

    def __str__(self):
        """
        Provide a human-readable string representation of the Resource instance.

        Returns:
            str: A string describing the resource's key attributes.
        """
        return f"Resource ID: {self.id} | Name: {self.name} | URL: {self.url} | Format : {self.format}"

    def _repr_html_(self):
        """
        Provide an HTML representation of the Resource instance for Jupyter display,
        with enhanced styles, watermark, and consistent formatting.
        """
        return resource_to_html(self)

    def open(self, mode="rb", **kwargs):
        """
        Open the resource for reading or writing.

        Args:
            mode (str): The mode in which to open the resource. This can be one of the
                following: 'r', 'rb', 'w', 'wb', 'a', 'ab'.

        Returns:
            file-like: A file-like object that can be used to read or write data.
        """
        client = client_for(self)
        url = self.url
        if not is_s3url(url):
            raise ValueError("Only s3 URLs are supported for path")
        return client.s3fs_open(url, mode=mode, **kwargs)

    def read_dataframe(self, format=None, **kwargs):
        """
        Read a DataFrame from the resource.

        Args:
            format (str): The format of the file to read. If not specified, the format will be
                inferred from the file extension.
            kwargs (dict): Additional keyword arguments to pass to the read.

        Returns:
            pd.DataFrame: The DataFrame read from the resource.
        """
        client = client_for(self)
        format = format or self.format or None
        return read_dataframe(client, self.url, format=format, **kwargs)

    def lineage(self, forward=False) -> Lineage:
        """
        Get the lineage of this resource.

        The lineage is a directed graph of resources and tasks that are reachable from this resource.
        Arcs in the graph correspond to resources being inputs or outputs of tasks.
        The lineage is used to track the flow of data and tasks in the STELAR system.

        The lineage can be either forward (from this resource to its descendants) or backward
        (from this resource to its ancestors).

        Args:
            forward (bool): If True, get the forward lineage; otherwise, get the backward lineage.

        Returns:
            Lineage: A Lineage object containing the lineage information.
        """
        from .lineage import Lineage

        linobj = api_call(self).resource_lineage(self.id, forward=forward)

        return Lineage(client_for(self), self.id, linobj, forward=forward)


class ResourceCursor(GenericCursor):
    """
    A cursor for a collection of STELAR resources.
    """

    def __init__(self, api):
        super().__init__(api, Resource)

    def fetch(self, **kwargs):
        raise NotImplementedError("ResourceCursor does not support fetch operations.")

    def fetch_list(self, **kwargs):
        raise NotImplementedError("ResourceCursor does not support fetch operations.")

    def search_url(self, path: str) -> ProxyVec:
        """Return resources whose url matches the given string.

        Args:
            path (str): The partial URL to search for.
        Returns:
            ProxyVec: A vector of resources with matching URLs.
        """
        res = self.search(query=[f"url:{path}"])
        rl = [UUID(r["id"]) for r in res["results"]]
        return ProxyVec(self.client, Resource, rl)

    def for_object(self, s3obj_spec: S3ObjSpec) -> ProxyVec:
        """Return resources for a given S3 object specification.

        Args:
            s3obj_spec (S3ObjSpec): The S3 object specification.
        Returns:
            ProxyVec: A vector of resources for the given S3 object
        """
        s3url = s3spec_to_url(s3obj_spec)
        return self.search_path(s3url)

    def search(
        self,
        *,
        query: list[str] | None = [],
        order_by: str = None,
        limit: int | None = None,
        offset: int | None = None,
    ):
        """
        Search for resources.

        This is the main function for searching the data catalog for resources.
        Other functions are implemented on top of this one.

        A resource query is a conunction of query terms of the form <field>:<match>.
        The query terms are combined with AND. The match is a string with optional wildcards.


        Arguments:
            query: A list of query strings. Each query string is of the form "<field>:<text>".
            order_by: A field to order by. Only a single field is allowed and only ascending
                     order is supported.
            limit: The maximum number of results to return.
            offset: The offset to start from.

        Returns:
            An answer dict which contains the following fields:
            - count: The number of results found (not the number of results returned).
            - results: A list of dicts, each corresponding to a resource entity.

        """
        search = self.client.api.get_call(Resource, "search")
        query = dict(
            query=query,
            order_by=order_by,
            limit=limit,
            offset=offset,
        )
        return search({k: v for k, v in query.items() if v is not None})



from uuid import UUID

import pytest


@pytest.fixture(autouse=True)
def tlicense(testcli):
    """Test fetching a license."""
    lic = testcli.licenses.create(
        title="Test License",
        key="test-license",
        description="A license for testing purposes.",
        url="https://example.com/test-license",
        osi_approved=True,
        open_data_approved=False,
    )

    yield lic.key

    try:
        lic.delete(purge=True)
    except Exception:
        pass


def test_create_license(testcli):
    try:
        lic = testcli.licenses.create(
            title="Test License",
            key="test-license2",
            description="A license for testing purposes.",
            url="https://example.com/test-license",
            osi_approved=True,
            open_data_approved=False,
        )
        """Test creating a license."""
        license = testcli.licenses.get("test-license")
        assert license is not None
        assert license.title == "Test License"
    finally:
        lic.delete()


def test_update_license(testcli):
    """Test updating a license."""
    license = testcli.licenses.get("test-license")
    assert license is not None

    license.title = "Updated Test License"
    license.description = "An updated description for testing purposes."

    updated_license = testcli.licenses.get("test-license")
    assert updated_license.title == "Updated Test License"
    assert updated_license.description == "An updated description for testing purposes."


def test_delete_license(testcli):
    """Test deleting a license."""
    license = testcli.licenses.get("test-license")
    assert license is not None
    license.delete()
    assert "test-license" not in testcli.licenses


def test_list_licenses(testcli):
    """Test listing all licenses."""
    assert "mit" in testcli.licenses
    assert "test-license" in testcli.licenses


def test_license_properties(testcli):
    """Test license properties."""
    license = testcli.licenses.get("test-license")
    assert license is not None

    assert isinstance(license.id, UUID)
    assert isinstance(license.key, str)
    assert isinstance(license.title, str)
    assert isinstance(license.description, str)
    assert isinstance(license.url, str)
    assert isinstance(license.osi_approved, bool)
    assert isinstance(license.open_data_approved, bool)


def test_license_proxy(testcli):
    """Test the LicensedProxy functionality."""
    license = testcli.licenses.get("test-license")
    assert license is not None


def test_dataset_license(testcli):
    """Test setting a license on a dataset."""

    d = testcli.datasets["dataset1"]

    # Clear the license
    d.license_id = None
    assert d.license_id is None
    assert d.license is None

    # Set the license to a string via license_id
    d.license_id = "test-license"
    assert d.license_id == "test-license"
    assert d.license.key == "test-license"

    # set the license via license
    d.license = "mit"
    assert d.license_id == "mit"
    assert d.license.key == "mit"

    # set via object
    d.license = testcli.licenses["test-license"]
    assert d.license_id == "test-license"
    assert d.license.key == "test-license"

    # set via UUID
    mitid = str(testcli.licenses["mit"].id)

    d.license = mitid
    assert d.license_id == mitid
    assert d.license.key == "mit"


def test_dataset_license_unknown(testcli):
    """Test setting a license to None on a dataset."""
    d = testcli.datasets["dataset1"]

    # Clear the license
    d.license_id = "custom"
    assert d.license_id == "custom"
    assert d.license is None



from importlib.metadata import PackageNotFoundError, version  # pragma: no cover

try:
    # Change here if project is renamed and does not equal the package name
    dist_name = __name__
    __version__ = version(dist_name)
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

from .client import Client
from .dataset import Dataset
from .group import Group, Organization
from .license import License
from .lineage import Lineage
from .policy import Policy
from .process import Process
from .proxy import ProxyState, deferred_sync

# Include all proxy exceptions
from .proxy.exceptions import *
from .proxy.exceptions import __all__ as __all_exceptions
from .relationship import Rel, Relationship, Relationships
from .resource import Resource
from .task_spec import TaskSpec
from .tasks import Task
from .tool import Image, ImageRegistryToken, Tool
from .user import User

# Include all utility functions
from .utils import *
from .utils import __all__ as __all_utils
from .vocab import Tag, Vocabulary
from .workflows import Workflow

__all__ = [
    *__all_exceptions,
    *__all_utils,
    "Client",
    "Dataset",
    "Resource",
    "Organization",
    "Group",
    "ProxyState",
    "Tag",
    "License",
    "Vocabulary",
    "User",
    "Process",
    "Policy",
    "Image",
    "ImageRegistryToken",
    "Relationship",
    "Rel",
    "Relationships",
    "Task",
    "TaskSpec",
    "Workflow",
    "Tool",
    "deferred_sync",
    "Lineage",
    "ProxyState",
]

del __all_exceptions
del __all_utils

import os

if os.getenv("SPHINX_BUILD"):
    # This prevents duplicate documentation by sphinx
    del __all__



from __future__ import annotations

from dataclasses import KW_ONLY, dataclass, field
from datetime import datetime
from typing import TYPE_CHECKING
from uuid import UUID

if TYPE_CHECKING:
    from .client import Client
    from .proxy import Proxy
    from .resource import Resource


@dataclass
class ENode:
    id: UUID
    type: str
    lineage: "Lineage"
    incoming: list[ENode] = field(default_factory=list, init=False)
    outgoing: list[ENode] = field(default_factory=list, init=False)
    _ = KW_ONLY

    @property
    def ent(self) -> Proxy:
        """Return the proxy for this ENode."""
        return self.lineage.client.registry_for(self.type).fetch_proxy(self.id)


@dataclass
class EResource(ENode):
    def __post_init__(self):
        self.type = "Resource"

    _ = KW_ONLY
    package_id: UUID | None = None
    url: str | None = None
    label: str = ""


@dataclass
class ETask(ENode):
    def __post_init__(self):
        self.type = "Task"

    _ = KW_ONLY
    label: str = ""
    state: str = "unknown"
    image: str | None = None
    process_id: UUID | None = None
    start_date: datetime | None = None
    end_date: datetime | None = None


class Lineage:
    """
    Represents a lineage graph between resources and tasks in the Stelar system.

    A lineage is a collection of tasks and resources that share a common ancestry
    or descendant.
    """

    def __init__(
        self, client: Client, focus_id: UUID, data: dict, forward: bool = False
    ):
        self.client = client
        self.focus_id = focus_id
        self.data = data
        self.forward = forward

        self.nodes: list[ENode] = []
        self.node: dict[UUID, ENode] = {}
        self.focus_node = None
        for node in self.data["nodes"]:
            if node["type"] == "Resource":
                enode = EResource(
                    id=UUID(node["id"]),
                    type=node["type"],
                    lineage=self,
                    label=node.get("label", ""),
                )
            elif node["type"] == "Task":
                enode = ETask(
                    id=UUID(node["id"]),
                    type=node["type"],
                    lineage=self,
                    label=node.get("label"),
                    state=node["state"],
                    image=node["image"],
                    start_date=datetime.fromisoformat(node["start_date"]),
                    end_date=datetime.fromisoformat(node["end_date"]),
                )
            else:
                raise ValueError(f"Unknown node type: {node['type']}")

            self.nodes.append(enode)
            self.node[enode.id] = enode
            if node.get("final"):
                self.focus_node = enode

        for s, t in self.data["edges"]:
            source = self.node[UUID(s)]
            target = self.node[UUID(t)]
            source.outgoing.append(target)
            target.incoming.append(source)

        self.resources: list[EResource] = [
            n for n in self.nodes if isinstance(n, EResource)
        ]
        self.tasks: list[ETask] = [n for n in self.nodes if isinstance(n, ETask)]

    @property
    def focus(self) -> Resource:
        """
        Return the focus resource of this lineage.

        The focus is the resource that is at the center of the lineage graph.
        """
        return self.client.resources.get(self.focus_id)

    def __repr__(self):
        return f"Lineage(focus={self.focus_id},forward={self.forward})"



from __future__ import annotations

from typing import TYPE_CHECKING
from uuid import UUID

from .generic import GenericCursor, GenericProxy
from .proxy import DateField, Id, NameId, Property, Proxy
from .proxy.fieldvalidation import BoolField, LicenseNameField, StrField
from .utils import client_for

if TYPE_CHECKING:
    from . import Client


class License(GenericProxy):
    id = Id()
    key = NameId(validator=LicenseNameField, entity_name="key")
    metadata_created = Property(validator=DateField)
    metadata_modified = Property(validator=DateField)

    title = Property(validator=StrField(nullable=False), updatable=True)
    url = Property(validator=StrField(nullable=True), updatable=True)
    description = Property(validator=StrField(nullable=True), updatable=True)
    image_url = Property(validator=StrField(nullable=True), updatable=True)

    osi_approved = Property(
        validator=BoolField(nullable=False, default=False), updatable=True
    )
    open_data_approved = Property(
        validator=BoolField(nullable=False, default=False), updatable=True
    )


class LicenseCursor(GenericCursor[License]):
    """Cursor for Licenses."""

    def __init__(self, client: Client):
        super().__init__(client, License)


class LicensedProxy(Proxy, entity=False):
    """A proxy that has a license."""

    license_id = Property(validator=StrField(nullable=True), updatable=True)

    @property
    def license(self) -> License | None:
        """Return the License object associated with this proxy, or none."""
        if self.license_id is None:
            return None
        return client_for(self).licenses.get(self.license_id)

    @license.setter
    def license(self, license: License | str | None):
        """Set the license for this proxy."""
        if license is None:
            self.license_id = None
        elif isinstance(license, License):
            self.license_id = license.key
        elif isinstance(license, str):
            self.license_id = license
        elif isinstance(license, UUID):
            self.license_id = str(license)
        else:
            raise TypeError(f"Unsupported type for license: {type(license)}")



from uuid import uuid4

from stelar.client import Dataset, Process, Rel, Relationship, Relationships, Workflow


def test_rel_enum():
    """Test the Rel enum."""

    for rel in Rel:
        assert isinstance(rel, Rel)
        assert isinstance(rel.value, str)
        assert rel.name.lower() == rel.value

    for rel in Rel:
        assert rel.peer().peer() is rel
        assert rel.is_canonical() or rel.peer().is_canonical()


def test_relationship_class(testcli):
    c = testcli

    iris = c.datasets.get("iris")
    wine = c.datasets.get("wine")

    r = Relationship(
        c,
        {
            "subject": str(iris.id),
            "subject_name": iris.name,
            "subject_type": "dataset",
            "object": str(wine.id),
            "object_name": wine.name,
            "object_type": "dataset",
            "relationship": Rel.LINKS_TO.value,
            "comment": "This is a test relationship.",
        },
    )

    assert r.subject_id == iris.id
    assert r.object_id == wine.id
    assert r.subject_name == iris.name
    assert r.object_name == wine.name
    assert r.subject_type == "dataset"
    assert r.object_type == "dataset"
    assert r.relationship is Rel.LINKS_TO

    assert Relationship.from_triple(iris, Rel.LINKS_TO, wine) == r

    assert r.subject is iris
    assert r.object is wine

    assert r.subject_data() == {
        "id": iris.id,
        "name": iris.name,
        "type": "dataset",
    }
    assert r.object_data() == {
        "id": wine.id,
        "name": wine.name,
        "type": "dataset",
    }

    # Check the relationship exists in the database.
    assert r.exists()
    assert r  # Cast to boolean

    rpeer = r.peer()
    assert rpeer.subject_id == wine.id
    assert rpeer.object_id == iris.id
    assert rpeer.subject_name == wine.name
    assert rpeer.object_name == iris.name
    assert rpeer.subject_type == "dataset"
    assert rpeer.object_type == "dataset"
    assert rpeer.relationship is Rel.LINKED_FROM

    # Check the peer relationship exists in the database.
    assert rpeer.exists()

    assert r == rpeer
    assert rpeer.subject is wine
    assert rpeer.object is iris
    assert rpeer.canonical() == r
    assert rpeer.canonical().relationship is Rel.LINKS_TO


def test_relationship_not_exists(testcli):
    c = testcli

    iris = c.datasets.get("iris")
    wine = c.datasets.get("wine")

    r = Relationship.from_triple(iris, Rel.LINKS_TO, wine)
    assert r.exists()

    r.relationship = Rel.CHILD_OF
    assert not r.exists()

    r.subject_id = uuid4()
    r.name = "Nonexistent1"

    assert not r.exists()
    assert not r


def test_relationship_match(testcli):
    c = testcli

    iris = c.datasets.get("iris")
    wine = c.datasets.get("wine")
    proc = c.processes.get("simple_proc")

    r = Relationship.from_triple(iris, Rel.PARENT_OF, proc)

    assert r.matches(iris, Rel.PARENT_OF, proc)
    assert r.matches(iris, "parent_of", proc)
    assert r.matches(iris, None, proc)
    assert r.matches(iris, Rel.PARENT_OF, None)
    assert r.matches(iris, None, None)
    assert r.matches(None, Rel.PARENT_OF, Process)
    assert r.matches(Process, Rel.CHILD_OF, None)
    assert r.matches(None, None, None)

    # Check reverse relationship
    assert r.matches(proc, Rel.CHILD_OF, iris)
    assert r.matches(proc, "child_of", iris)
    assert r.matches(proc, None, iris)
    assert r.matches(proc, Rel.CHILD_OF, None)
    assert r.matches(proc, None, None)
    assert r.matches(None, Rel.CHILD_OF, Dataset)
    assert r.matches(Dataset, Rel.PARENT_OF, None)
    assert r.matches(None, Rel.PARENT_OF, None)
    assert r.matches(None, Rel.CHILD_OF, None)

    # Check non-matching cases
    assert not r.matches(iris, Rel.LINKS_TO, proc)
    assert not r.matches(iris, Rel.PARENT_OF, wine)
    assert not r.matches(wine, Rel.PARENT_OF, iris)
    assert not r.matches(Workflow, None, None)
    assert not r.matches(None, Rel.DEPENDENCY_OF, None)


def test_relationships(testcli):
    c = testcli

    # use test datasets
    dset1 = c.datasets.get("dataset1")
    dset2 = c.datasets.get("dataset2")
    dset3 = c.datasets.get("dataset3")
    proc = c.processes.get("simple_proc")

    # Create a relationship
    dset1.relationships.add(Rel.PARENT_OF, dset2, "Test relationship 1")
    dset1.relationships.add(Rel.PARENT_OF, dset3, "Test relationship 2")
    dset2.relationships.add(Rel.DEPENDS_ON, dset3, "Test relationship 3")
    dset3.add_relationship("depends_on", proc, "Test relationship 4")

    rs1 = dset1.relationships()
    rs2 = dset2.relationships()
    rs3 = dset3.relationships()
    rp = proc.relationships()

    assert len(rs1) == 2
    assert len(rs2) == 2
    assert len(rs3) == 3
    assert len(rp) == 1

    rall = rs1 | rs2 | rs3 | rp
    assert len(rall) == 4

    try:
        # check membership
        assert all(r in rall for r in rs1)
        assert all(r in rall for r in rs2)
        assert all(r in rall for r in rs3)
        assert all(r in rall for r in rp)

        # check relationships set operations
        assert rs1.objects() == {dset2, dset3}
        assert rs3.entities() == {dset1, dset2, dset3, proc}
        assert rs1.entities() == {dset1, dset2, dset3}
        assert rp.objects() == {dset3}
        assert rp.entities() == {dset3, proc}

        assert (rs1 ^ rs2 ^ rs3 ^ rp) == Relationships(c, {})
        assert rall - rs3 == rs1 & rs2

        assert set() <= rs3 <= rall
    finally:
        rall.delete()


def test_proxy_relationship_get(testcli):
    c = testcli

    iris = c.datasets.get("iris")
    wine = c.datasets.get("wine")

    # There should be a relationship iris --links_to--> wine.
    # Check with the API.
    checking = c.GET("v2/relationships/iris/links_to/wine").json()
    assert checking["success"]
    assert checking["result"]
    # Same as above
    assert Relationship.from_triple(iris, Rel.LINKS_TO, wine).exists()

    # Check with the relationship object.
    rel = iris.relationships.get("links_to", "wine")

    assert iris.relationships.get("links_to", "wine") == rel
    assert wine.relationships.get(Rel.LINKED_FROM, "iris") == rel

    assert rel == iris.relationships.get(Rel.LINKS_TO, wine)
    assert rel.subject_id == iris.id
    assert rel.object_id == wine.id
    assert rel.subject_name == iris.name
    assert rel.object_name == wine.name
    assert rel.subject_type == "dataset"
    assert rel.object_type == "dataset"
    assert rel.relationship == Rel.LINKS_TO
